// 
// Copyright (c) 2009, Daisuke Deguchi, Masafumi Noda
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the Nagoya University nor the names of its contributors
// may be used to endorse or promote products derived from this software
// without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
// IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
// THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 

#include <iostream>
#include <mist/mist.h>
#include <mist/io/bmp.h>
#include <mist/timer.h>

// RX^g̐錾
__constant__ float coeff[ 512 ];

// eNX`̐錾
texture< float4, 1 > imgTexX;
texture< float4, 1 > imgTexY;

__global__ void kernel_x( float4 *out, int w, int h, int R )
{
	int i = threadIdx.x + blockDim.x * blockIdx.x;
	int j = threadIdx.y + blockDim.y * blockIdx.y;

	// ͈͓̉fɑ΂Ă̂ݏs
	if( i < w && j < h )
	{
		int s = i - R < 0  ? 0 - i : -R;
		int e = i + R >= w ? w - i :  R;

		// X̕
		float4 pix = make_float4( 0.0f, 0.0f, 0.0f, 0.0f );
		float  sum = 0.0f;
		for( int n = s ; n < e ; n++ )
		{
			float  g = coeff[ n + R ];
			float4 p = tex1Dfetch( imgTexX, i + n + j * w );

			pix.x += p.x * g;
			pix.y += p.y * g;
			pix.z += p.z * g;
			sum += g;
		}

		out[ i + j * w ] = make_float4( pix.x / sum, pix.y / sum, pix.z / sum, 255.0f );
	}
}
__global__ void kernel_y( float4 *out, int w, int h, int R )
{
	int i = threadIdx.x + blockDim.x * blockIdx.x;
	int j = threadIdx.y + blockDim.y * blockIdx.y;

	// ͈͓̉fɑ΂Ă̂ݏs
	if( i < w && j < h )
	{
		int s = j - R < 0  ? 0 - j : -R;
		int e = j + R >= h ? h - j :  R;

		// Y̕
		float4 pix = make_float4( 0.0f, 0.0f, 0.0f, 0.0f );
		float  sum = 0.0f;
		for( int n = s ; n < e ; n++ )
		{
			float g = coeff[ n + R ];
			float4 p = tex1Dfetch( imgTexY, i + ( j + n ) * w );

			pix.x += p.x * g;
			pix.y += p.y * g;
			pix.z += p.z * g;
			sum += g;
		}

		out[ i + j * w ] = make_float4( pix.x / sum, pix.y / sum, pix.z / sum, 255.0f );
	}
}

template < class Array >
inline void compute_gaussian_kernel( Array &kernel, double sigma )
{
	typedef typename Array::size_type		size_type;
	typedef typename Array::difference_type	difference_type;
	typedef typename Array::value_type		value_type;

	if( sigma == 0.0 )
	{
		sigma = 1.0;
	}

	double radius = sigma * 2.0;

	difference_type R = static_cast< difference_type >( std::ceil( radius ) );

	kernel.resize( R * 2 + 1 );

	difference_type fw = kernel.size( );
	difference_type rw = fw / 2;

	double _1_sig2 = 1.0 / ( sigma * sigma * 2.0 );
	double sum = 0.0;
	for( difference_type x = -rw ; x <= rw ; x++ )
	{
		double xx = static_cast< double >( x * x );
		double g = std::exp( -xx * _1_sig2 );
		sum += g;
		kernel[ x + rw ] = static_cast< value_type >( g );
	}

	for( size_type i = 0 ; i < kernel.size( ) ; i++ )
	{
		kernel[ i ] /= static_cast< value_type >( sum );
	}
}

int main( int argc, char *argv[] )
{
	// ̓t@Cw肳Ă邩ǂ`FbN
	if( argc != 4 )
	{
		printf( "This program requires path of a input and a output images with a parameter of sigma.\n" );
		return( 1 );
	}

	// ͉摜IuWFNgp̕ϐ
	mist::array2< mist::rgba< float > > img;
	mist::array< float >  cf;

	// KEVAJ[lvZ
	compute_gaussian_kernel( cf, atof( argv[ 3 ] ) );

	// 摜ǂݍ
	if( !mist::read_bmp( img, argv[ 1 ] ) )
	{
		printf( "Failed to read image from [%s].\n", argv[ 1 ] );
		return( 1 );
	}

	// GPUɓ͉摜p̃̈mۂCeNX`Ƀ}bsO
	cudaMemcpyToSymbol( coeff, &cf[ 0 ], cf.size( ) * sizeof( float ) );

	// GPȔo͉摜IuWFNgp̕ϐ
	float4 *iData = NULL, *oData = NULL, *hData = NULL;
	cudaMallocHost( ( void ** )&hData, img.size( ) * sizeof( float4 ) );
	cudaMalloc( ( void ** )&iData, img.size( ) * sizeof( float4 ) );
	cudaMalloc( ( void ** )&oData, img.size( ) * sizeof( float4 ) );
	cudaMemcpy( oData, &img[ 0 ], img.size( ) * sizeof( float4 ), cudaMemcpyHostToDevice );

	cudaBindTexture( 0, imgTexX, oData );
	cudaBindTexture( 0, imgTexY, iData );

	{
		mist::timer t;

		int nloops = 100;
		for( int loop = 0 ; loop < nloops ; loop++ )
		{
			// sGPŨXbhw
			int threadNumX = 16;
			int threadNumY = 16;
			int blockNumX = ( int )img.width( )  / threadNumX + ( ( img.width( )  % threadNumX ) == 0 ? 0 : 1 );
			int blockNumY = ( int )img.height( ) / threadNumY + ( ( img.height( ) % threadNumY ) == 0 ? 0 : 1 );
			dim3 nThreads( threadNumX, threadNumY, 1 );
			dim3 nBlocks( blockNumX, blockNumY, 1 );

			// GPUX̕s
			kernel_x<<< nBlocks, nThreads >>>( iData, img.width( ), img.height( ), cf.size( ) / 2 );

			// ܂őҋ@
			cudaThreadSynchronize( );

			// GPUX̕s
			kernel_y<<< nBlocks, nThreads >>>( oData, img.width( ), img.height( ), cf.size( ) / 2 );

			// GPȔʂ]
			cudaMemcpy( hData, oData, img.size( ) * sizeof( float4 ), cudaMemcpyDeviceToHost );
		}

		printf( "Computation time: %.3f msec\n", t.elapse( ) * 1000 / nloops );
	}

	// ʂ̉摜ۑ
	mist::array2< mist::rgba< float > > out( img.width( ), img.height( ), ( mist::rgba< float > * )hData, img.size( ) * sizeof( float4 ) );
	if( !mist::write_bmp( out, argv[ 2 ] ) )
	{
		printf( "Failed to save image to [%s].\n", argv[ 2 ] );
		return( 1 );
	}

	// ̌n
	cudaUnbindTexture( imgTexX );
	cudaUnbindTexture( imgTexY );
	cudaFreeHost( hData );
	cudaFree( iData );
	cudaFree( oData );

	return( 0 );
}
