// 
// Copyright (c) 2009, Daisuke Deguchi, Masafumi Noda
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the Nagoya University nor the names of its contributors
// may be used to endorse or promote products derived from this software
// without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
// IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
// THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <mist/timer.h>

__host__ void multiply0( float *hA, float *hB, float *hC, int N )
{
	for( int m = 0 ; m < N ; m++ )
	{
		for( int n = 0 ; n < N ; n++ )
		{
			float sum = 0.0f;
			for( int k = 0 ; k < N ; k++ )
			{
				sum += hA[ m + k * N ] * hB[ k + n * N ];
			}

			hC[ m + n * N ] = sum;
		}
	}
}

__global__ void multiply1( float *dA, float *dB, float *dC, int N )
{
	int n = threadIdx.x + blockIdx.x * blockDim.x;
	int m = threadIdx.y + blockIdx.y * blockDim.y;

	float sum = 0.0f;
	for( int k = 0 ; k < N ; k++ )
	{
		sum += dA[ m + k * N ] * dB[ k + n * N ];
	}

	dC[ m + n * N ] = sum;
}

__global__ void multiply2( float *dA, float *dB, float *dC, int N )
{
	int n = threadIdx.x + blockIdx.x * blockDim.x;
	int m = threadIdx.y + blockIdx.y * blockDim.y;

	float sum = 0.0f;
	for( int k = 0 ; k < N ; k += 16 )
	{
        __shared__ float tA[ 16 ][ 16 ];
        __shared__ float tB[ 16 ][ 16 ];

		tA[ threadIdx.y ][ threadIdx.x ] = dA[ m + ( k + threadIdx.x ) * N ];
		tB[ threadIdx.x ][ threadIdx.y ] = dB[ ( k + threadIdx.y ) + n * N ];

		__syncthreads( );

		for( int t = 0 ; t < 16 ; t++ )
		{
			sum += tA[ threadIdx.y ][ t ] * tB[ threadIdx.x ][ t ];
		}

		__syncthreads( );
	}

	dC[ m + n * N ] = sum;
}

void print_matrix( float *A, int N )
{
	for( int m = 0 ; m < N ; m++ )
	{
		for( int n = 0 ; n < N ; n++ )
		{
			printf( "%.1f, ", A[ m + n * N ] );
		}
		printf( "\n" );
	}

	printf( "\n" );
}


int main( int argc, char *argv[] )
{
	int N = 512;			// sA, B, C̃TCY
	float *hA, *hB, *hC;	// CPUŗp郁ւ̃|C^
	float *dA, *dB, *dC;	// GPUŗp郁ւ̃|C^

	/* CPŨm */
	hA = ( float * )malloc( N * N * sizeof( float ) );
	hB = ( float * )malloc( N * N * sizeof( float ) );
	hC = ( float * )malloc( N * N * sizeof( float ) );

	/* GPŨm */
	cudaMalloc( ( void ** )&dA, N * N * sizeof( float ) );
	cudaMalloc( ( void ** )&dB, N * N * sizeof( float ) );
	cudaMalloc( ( void ** )&dC, N * N * sizeof( float ) );

	/* ōs̊evfɒlݒ */
	srand( static_cast< unsigned int >( time( NULL ) ) );
	for( int i = 0 ; i < N * N ; i++ )
	{
		hA[ i ] = rand( ) / static_cast< float >( RAND_MAX  );
		hB[ i ] = rand( ) / static_cast< float >( RAND_MAX );
	}

	/* CPŨGPU֓] */
	cudaMemcpy( dA, hA, N * N * sizeof( float ), cudaMemcpyHostToDevice );
	cudaMemcpy( dB, hB, N * N * sizeof( float ), cudaMemcpyHostToDevice );

	/* CPUgꍇ̌vZԂ] */
	{
		// vZԂ̌vJn
		//mist::timer t;

		for( int loop = 0 ; loop < 10 ; loop++ )
		{
			multiply0( hA, hB, hC, N );
		}

		//printf( "Computation time (CPU): %.1f msec.\n", t.elapse( ) * 100.0 );
	}

	/* vZ hC ̒lŊmF */
	//print_matrix( hC, N );

	/* GPUgꍇ̌vZԂv */
	for( int mode = 0 ; mode < 2 ; mode++ )
	{
		// vZԂ̌vJn
		//mist::timer t;

		for( int loop = 0 ; loop < 10 ; loop++ )
		{
			/* sGPŨXbhCubNݒ */
			dim3 nThreads( 16, 16 );
			dim3 nBlocks( N / nThreads.x, N / nThreads.y );

			/* GPŨJ[lsCC=A~B ̌ʂ dC Ɋi[ */
			if( mode == 0 )
			{
				multiply1<<< nBlocks, nThreads >>>( dA, dB, dC, N );
			}
			else
			{
				multiply2<<< nBlocks, nThreads >>>( dA, dB, dC, N );
			}

			/* GPǓvZʂCPU֓] */
			cudaMemcpy( hC, dC, N * N * sizeof( float ), cudaMemcpyDeviceToHost );
		}

		//printf( "Computation time (kernel%d): %.1f msec.\n", mode, t.elapse( ) * 100.0 );
	}

	/* vZ hC ̒lŊmF */
	//print_matrix( hC, N );

	/* CPUGPUꂼ̃ */
	cudaFree( dA );
	cudaFree( dB );
	cudaFree( dC );
	free( hA );
	free( hB );
	free( hC );

	return( 0 );
}
